/*
 * Copyright 2005 Ivan Leo Puoti
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdarg.h>

#include "windef.h"
#include "winbase.h"
#include "winreg.h"
#include "winternl.h"
#include "wine/debug.h"
#include "wine/unicode.h"
#include "winnt.h"
#include "basetsd.h"
#include "ntstatus.h"
#include "winioctl.h"
#include "ntoskrnl.h"
#include "wine/ntoskrnl_protocol.h"

/* Driver calls */
typedef NTSTATUS (WINAPI *DRIVERENTRYPOINT)(PDRIVER_OBJECT, PUNICODE_STRING);
typedef NTSTATUS (WINAPI *UNLOAD)(PDRIVER_OBJECT);
/* Driver managment calls */
static NTSTATUS load_driver(PWSTR);
static NTSTATUS get_handle(PWSTR DriverData, HANDLE pipe);
static NTSTATUS unload_driver(PWSTR DriverData);
static NTSTATUS driver_control(PWSTR DriverData, HANDLE pipe);
static void add_driver(PDRIVER_OBJECT driver, PWSTR key, HMODULE DriverImage);
static int remove_driver(PDRIVER_OBJECT driver);
static VOID WINAPI DriverManagment(HANDLE pipe);

WINE_DEFAULT_DEBUG_CHANNEL(ntoskrnl); 
DWORD NtBuildNumber = 0xf0000893;
UNICODE_STRING HardwareDB;
DRIVERDATABASE *DriverDatabase;

INT WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, INT nCmdShow)
{
    HANDLE pipe;
    static const WCHAR PipeName[] = {'\\','\\','.','\\','p','i','p','e','\\','n','t','o','s','k','r','n','l',0};
    static const WCHAR HardwareDBString[] = {'\\','R','e','g','i','s','t','r','y','\\',
                                             'M','a','c','h','i','n','e','\\',
                                             'H','a','r','d','w','a','r','e',0};
    BOOL connect;

    DriverDatabase = RtlAllocateHeap(GetProcessHeap(), 0, sizeof(DRIVERDATABASE));
    if((NTSTATUS)DriverDatabase == STATUS_NO_MEMORY || (NTSTATUS)DriverDatabase == STATUS_ACCESS_VIOLATION)
        return FALSE;
   
    RtlZeroMemory(DriverDatabase, sizeof(DRIVERDATABASE));

    RtlCreateUnicodeString(&HardwareDB, HardwareDBString);
    for(;;)
    {
        pipe = CreateNamedPipeW(PipeName, PIPE_ACCESS_DUPLEX, PIPE_TYPE_BYTE | PIPE_READMODE_BYTE | PIPE_WAIT,
                                PIPE_UNLIMITED_INSTANCES, BUFFER, BUFFER, NMPWAIT_USE_DEFAULT_WAIT, NULL);
        if(pipe == INVALID_HANDLE_VALUE)
        {
            WINE_ERR("CreateNamedPipe failed with %ld\n", GetLastError());
            return FALSE;
        }

        /* According to MSDN if ERROR_PIPE_CONNECTED is returned by GetLastError when the pipe client
           attempted to connect between the CreateNamedPipe and the ConnectNamedPipe calls,
           the connection is good even if the function returns zero */
        if(!(connect = ConnectNamedPipe(pipe, NULL)) && GetLastError() != ERROR_PIPE_CONNECTED)
        {
            CloseHandle(pipe);
            return FALSE;
        }
        DriverManagment(pipe);
        if(DriverDatabase == NULL)
            goto end;
    }
    end:
    return TRUE;
}

static VOID WINAPI DriverManagment(HANDLE pipe)
{
    NTOSKRNL_CALL call;
    PWSTR DriverData = NULL;
    NTSTATUS retv = STATUS_SUCCESS;
    IO_STATUS_BLOCK StatusBlock;

    retv = NtReadFile(pipe, NULL, NULL, NULL, &StatusBlock, &call, sizeof(NTOSKRNL_CALL), 0, NULL);
    if(retv != STATUS_SUCCESS)
        goto error;

    if(call.size > 0)
    {
        DriverData = (PWSTR)RtlAllocateHeap(GetProcessHeap(), 0, (call.size)*sizeof(WCHAR)+sizeof(WCHAR));
        retv = NtReadFile(pipe, NULL, NULL, NULL, &StatusBlock, DriverData, call.size, 0, NULL);
        if(retv != STATUS_SUCCESS)
        {
            WINE_ERR("NtReadFile failed to open pipe, error %lx\n", retv);
            goto error;
        }
        DriverData[call.size] = 0;
    }

    switch(call.type){
    case LOADDRIVER :
        retv = load_driver(DriverData);
        break;
    case GETDEVICEHANDLE :
        retv = get_handle(DriverData, pipe);
        if(retv == STATUS_SUCCESS)
            return;
        break;
    case UNLOADDRIVER :
        retv = unload_driver(DriverData);
        break;
    case IOCONTROL :
        retv = driver_control(DriverData, pipe);
    }
    error:
    RtlFreeHeap(GetProcessHeap(), 0, DriverData);
    NtWriteFile(pipe, NULL, NULL, NULL, &StatusBlock, &retv, sizeof(NTSTATUS), 0, NULL);
    NtFlushBuffersFile(pipe, &StatusBlock);
    DisconnectNamedPipe(pipe);
    NtClose(pipe);
}

static NTSTATUS load_driver(PWSTR DriverData)
{
    static const WCHAR ValidPath[] = {'\\','R','e','g','i','s','t','r','y','\\','M','a','c','h','i','n','e','\\',
           'S','y','s','t','e','m','\\','C','u','r','r','e','n','t','C','o','n','t','r','o','l','S','e','t','\\',
           'S','e','r','v','i','c','e','s','\\',0};
    static const WCHAR type[] = {'T','y','p','e',0};
    static const WCHAR ImagePath[] = {'I','m','a','g','e','P','a','t','h',0};
    PDRIVER_OBJECT pDriverObject;
    UNICODE_STRING ValidPathString, Key, PassedString;;
    PKEY_VALUE_PARTIAL_INFORMATION Info = NULL;
    ULONG length;
    INT ValidStringCheck;
    OBJECT_ATTRIBUTES object;
    HKEY DriverKeyHandle;
    HMODULE ViewAddress = NULL;
    PIMAGE_NT_HEADERS NtHeader;
    NTSTATUS retv;

    /* Check if we have a valid path */
    RtlCreateUnicodeString(&ValidPathString, ValidPath);
    PassedString.Buffer = RtlAllocateHeap(GetProcessHeap(), 0, (strlenW(ValidPath)+1)*sizeof(WCHAR));
    memcpy(PassedString.Buffer, DriverData, sizeof(WCHAR)*(strlenW(ValidPath)+1));
    PassedString.Buffer[strlenW(ValidPath)] = 0;
    PassedString.Length = strlenW(PassedString.Buffer)*sizeof(WCHAR);
    ValidStringCheck = RtlCompareUnicodeString(&ValidPathString, &PassedString, TRUE);
    RtlFreeHeap(GetProcessHeap(), 0, ValidPathString.Buffer);
    RtlFreeHeap(GetProcessHeap(), 0, PassedString.Buffer);
    if(ValidStringCheck != 0)
        return STATUS_OBJECT_PATH_SYNTAX_BAD;

    /* check if the key exists */
    if(RtlCheckRegistryKey(RTL_REGISTRY_ABSOLUTE, DriverData) != STATUS_SUCCESS)
        return STATUS_OBJECT_NAME_NOT_FOUND;
       
    /* Check we have a SERVICE_KERNEL_DRIVER service */
    RtlCreateUnicodeString(&ValidPathString, DriverData);
    InitializeObjectAttributes(&object, &ValidPathString, 0, NULL, NULL);
    NtOpenKey(&DriverKeyHandle, KEY_READ, &object);
    RtlCreateUnicodeString(&Key, type);
    NtQueryValueKey(DriverKeyHandle, &Key, KeyValuePartialInformation, Info, 0, &length);
    Info = RtlAllocateHeap(GetProcessHeap(), 0, length);
    NtQueryValueKey(DriverKeyHandle, &Key, KeyValuePartialInformation, Info, length, &length);
    if(*Info->Data != SERVICE_KERNEL_DRIVER)
    {
        NtClose(DriverKeyHandle);
        RtlFreeHeap(GetProcessHeap(), 0, Key.Buffer);
        RtlFreeHeap(GetProcessHeap(), 0, Info);
        return STATUS_NOT_IMPLEMENTED; /* and probably never will be */
    }
    RtlFreeHeap(GetProcessHeap(), 0, Info);

    /* Get path to driver image */
    RtlInitUnicodeString(&ValidPathString, ImagePath);
    NtQueryValueKey(DriverKeyHandle, &ValidPathString, KeyValuePartialInformation, Info, 0, &length);
    Info = RtlAllocateHeap(GetProcessHeap(), 0, length);
    NtQueryValueKey(DriverKeyHandle, &ValidPathString, KeyValuePartialInformation, Info, length, &length);

    /* Check the driver isn't already loaded */
    if(find_driver(NULL, DriverData, NULL, NULL, NULL) != NULL)
        return STATUS_IMAGE_ALREADY_LOADED;

    /* Map file to memory */
    RtlCreateUnicodeString(&PassedString, (PCWSTR)Info->Data);
    NtClose(DriverKeyHandle);
    retv = LdrLoadDll(NULL, 0, &PassedString, &ViewAddress);
    RtlFreeHeap(GetProcessHeap(), 0, PassedString.Buffer);
    if(retv != STATUS_SUCCESS)
        return STATUS_DRIVER_UNABLE_TO_LOAD;

    /* Get entry point */
    NtHeader = RtlImageNtHeader(ViewAddress);
    if(NtHeader == NULL)
        return STATUS_DRIVER_ENTRYPOINT_NOT_FOUND;

    /* Initialise DriverObject */
    pDriverObject = RtlAllocateHeap(GetProcessHeap(), 0, sizeof(DRIVER_OBJECT));
    pDriverObject->HardwareDatabase = &HardwareDB;
    pDriverObject->Size = sizeof(DRIVER_OBJECT); /* Wild guess */
    pDriverObject->DriverInit = (DRIVERENTRYPOINT)((char*)ViewAddress + NtHeader->OptionalHeader.AddressOfEntryPoint);

    /* Add driver to database */
    pDriverObject->DeviceObject = NULL;
    add_driver(pDriverObject, DriverData, ViewAddress);

    /* Call DriverEntry */
    RtlInitUnicodeString(&ValidPathString, DriverData);
    retv = ((DRIVERENTRYPOINT)(pDriverObject->DriverInit))(pDriverObject, &ValidPathString);
    RtlFreeHeap(GetProcessHeap(), 0, ValidPathString.Buffer);
    if (retv != STATUS_SUCCESS)
    {
        WINE_ERR("DriverEntry returned %lx\n", retv);
        LdrUnloadDll(ViewAddress);
        remove_driver(pDriverObject);
        retv = STATUS_FAILED_DRIVER_ENTRY;
    }
    return retv;
}

static NTSTATUS get_handle(PWSTR DriverData, HANDLE pipe)
{
    UNICODE_STRING PassedString;
    PDRIVERDATABASE LocalDB = NULL;
    NTSTATUS retv;
    IO_STATUS_BLOCK StatusBlock;

    RtlCreateUnicodeString(&PassedString, DriverData);
    LocalDB = find_driver(NULL, NULL, NULL, NULL, &PassedString);
    RtlFreeHeap(GetProcessHeap(), 0, PassedString.Buffer);
    if(!LocalDB)
        return STATUS_NO_SUCH_DEVICE;
    else
    {
        retv = STATUS_SUCCESS;
        NtWriteFile(pipe, NULL, NULL, NULL, &StatusBlock, &retv, sizeof(NTSTATUS), 0, NULL);
        NtWriteFile(pipe, NULL, NULL, NULL, &StatusBlock, &LocalDB, sizeof(DWORD), 0, NULL);
        NtFlushBuffersFile(pipe, &StatusBlock);
        DisconnectNamedPipe(pipe);
        NtClose(pipe);
        return retv;
    }
}

static NTSTATUS unload_driver(PWSTR DriverData)
{
    PDRIVERDATABASE LocalDB = NULL;

    LocalDB = find_driver(NULL, DriverData, NULL, NULL, NULL);
    if(LocalDB)
    {
        ((UNLOAD)(LocalDB->Driver->DriverUnload))(LocalDB->Driver);
        LdrUnloadDll(LocalDB->DriverImage);
        remove_driver(LocalDB->Driver);
        return STATUS_SUCCESS;
    }
    else
        return STATUS_OBJECT_NAME_NOT_FOUND;
}

static NTSTATUS driver_control(PWSTR DriverData, HANDLE pipe)
{
    PDEVICE_OBJECT Device = NULL;
    IRP Irp;
    IO_STACK_LOCATION Sp;
    IOCONTROL_CALL IoCall;
    IOCONTROL_RET  IoRet;
    DWORD* Buffer = NULL;
    IO_STATUS_BLOCK StatusBlock;
    NTSTATUS retv;

    NtReadFile(pipe, NULL, NULL, NULL, &StatusBlock, &IoCall, sizeof(IOCONTROL_CALL), 0, NULL);
    Device = IoCall.Handle;

    /* Initialise Irp */
    WINE_TRACE("Handling IoContorl, handle is %p, ControlCode is %lx,OutputBufferLength is %lx,"
               "InputBufferLength is %lx\n", IoCall.Handle, IoCall.IoControl,
               IoCall.OutputBufferLength, IoCall.InputBufferLength);
    Irp.Tail.Overlay.CurrentStackLocation = &Sp;

    Sp.Parameters.DeviceIoControl.OutputBufferLength = IoCall.OutputBufferLength;
    Sp.Parameters.DeviceIoControl.InputBufferLength = IoCall.InputBufferLength;
    Sp.Parameters.DeviceIoControl.IoControlCode = IoCall.IoControl;
    Sp.MajorFunction = IRP_MJ_DEVICE_CONTROL;

    /* METHOD_BUFFERED */
    if(IO_METHOD_FROM_CTL_CODE(IoCall.IoControl) == METHOD_BUFFERED)
    {
        Irp.AssociatedIrp.SystemBuffer = RtlAllocateHeap(GetProcessHeap(), 0, IoCall.InputBufferLength < IoCall.OutputBufferLength ?  IoCall.InputBufferLength : IoCall.OutputBufferLength);
        NtReadFile(pipe, NULL, NULL, NULL, &StatusBlock, Irp.AssociatedIrp.SystemBuffer,
                   IoCall.OutputBufferLength, 0, NULL);
        Irp.MdlAddress = NULL;
        Buffer = Irp.AssociatedIrp.SystemBuffer;
    }

    /* METHOD_IN_DIRECT */
    if(IO_METHOD_FROM_CTL_CODE(IoCall.IoControl) == METHOD_IN_DIRECT)
    {
        if(IoCall.InputBufferLength > 0)
        {
            Irp.AssociatedIrp.SystemBuffer = RtlAllocateHeap(GetProcessHeap(), 0, IoCall.InputBufferLength);
            NtReadFile(pipe, NULL, NULL, NULL, &StatusBlock,Irp.AssociatedIrp.SystemBuffer,
                       IoCall.OutputBufferLength, 0, NULL);
        }
        else
            Irp.AssociatedIrp.SystemBuffer = NULL;
        /* FIXME: Allocate IoCall.OutputBufferLength bytes, point Irp.MdlAddress to it,
         * driver must have read access
         */
    }

    /* METHOD_OUT_DIRECT */
    if(IO_METHOD_FROM_CTL_CODE(IoCall.IoControl) == METHOD_OUT_DIRECT)
    {
        if(IoCall.InputBufferLength > 0)
        {
            Irp.AssociatedIrp.SystemBuffer = RtlAllocateHeap(GetProcessHeap(), 0, IoCall.InputBufferLength);
            NtReadFile(pipe, NULL, NULL, NULL, &StatusBlock, Irp.AssociatedIrp.SystemBuffer,
                       IoCall.OutputBufferLength, 0, NULL);
        }
       else
           Irp.AssociatedIrp.SystemBuffer = NULL;
        /* FIXME: Allocate IoCall.OutputBufferLength bytes, point Irp.MdlAddress to it,
         * Driver must have write access 
         */
    }

    /* METHOD_NEITHER */
    if(IO_METHOD_FROM_CTL_CODE(IoCall.IoControl) == METHOD_NEITHER)
    {
        Irp.UserBuffer = RtlAllocateHeap(GetProcessHeap(), 0, IoCall.OutputBufferLength);
        if(IoCall.InputBufferLength > 0)
        {
            Sp.Parameters.DeviceIoControl.Type3InputBuffer = RtlAllocateHeap(GetProcessHeap(),
                                                                             0, IoCall.InputBufferLength);
            NtReadFile(pipe, NULL, NULL, NULL, &StatusBlock,
                       Sp.Parameters.DeviceIoControl.Type3InputBuffer, IoCall.InputBufferLength, 0, NULL);
        }
        Buffer = Irp.UserBuffer;
    }
    Irp.RequestorMode = UserMode;
    Irp.Flags = 0;
    Irp.PendingReturned = FALSE;

    /* Call DispatchDeviceContorl and return */
    retv = ((PDRIVER_DISPATCH)(Device->DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]))(Device, &Irp);
    if(Irp.CancelIrql == TRUE && Irp.CancelRoutine != NULL)
        Irp.CancelRoutine(IoCall.Handle, &Irp);

    IoRet.IoStatus = Irp.IoStatus.Status;
    IoRet.Info = Irp.IoStatus.Information;
    if(IoRet.IoStatus != STATUS_SUCCESS)
        WINE_ERR("DispatchDeviceControl returned %lx\n", IoRet.IoStatus);
    NtWriteFile(pipe, NULL, NULL, NULL, &StatusBlock, &IoRet, sizeof(IOCONTROL_RET), 0, NULL);
    if(IoCall.OutputBufferLength > 0 && Buffer != NULL)
        NtWriteFile(pipe, NULL, NULL, NULL, &StatusBlock, Buffer, IoCall.OutputBufferLength, 0, NULL);
    return retv;
}

static void add_driver(PDRIVER_OBJECT driver, PWSTR key, HMODULE DriverImage)
{
    PDRIVERDATABASE LocalDB = DriverDatabase;
    while(LocalDB->down != NULL)
        LocalDB = LocalDB->down;
    if(LocalDB->up == NULL)
    {
        LocalDB->Driver = driver;
        LocalDB->Key = RtlAllocateHeap(GetProcessHeap(), 0, (strlenW(key)+1)*sizeof(WCHAR));
        LocalDB->Key = key;
        LocalDB->DriverImage = DriverImage;
    }
    else
    {
        LocalDB->down = RtlAllocateHeap(GetProcessHeap(), 0, sizeof(DRIVERDATABASE));
        LocalDB->down->up = LocalDB;
        RtlZeroMemory(LocalDB, sizeof(DRIVERDATABASE));
        LocalDB = LocalDB->down;
        LocalDB->Driver = driver;
        LocalDB->Key = RtlAllocateHeap(GetProcessHeap(), 0, (strlenW(key)+1)*sizeof(WCHAR));
        LocalDB->Key = key;
        LocalDB->DriverImage = DriverImage;
        LocalDB->down = NULL;
    }
}

static int remove_driver(PDRIVER_OBJECT driver)
{
    int found = FALSE;
    PDRIVERDATABASE LocalDB;
    LocalDB = find_driver(driver, NULL, NULL, NULL, NULL);
    if(LocalDB == NULL)
        return FALSE;
    while(found == FALSE)
    {
        if(LocalDB->Driver == driver)
        {
            found = TRUE;
            break;
        }
        if(LocalDB->down != NULL)
            LocalDB = LocalDB->down;
        else
            break;
    }
    if(found == TRUE)
    {
        if(LocalDB->down != NULL)
        {
            if(LocalDB->down->up != NULL && LocalDB->up != NULL)
                LocalDB->down->up = LocalDB->up;
        }
        if(LocalDB->up != NULL)
        {
            if(LocalDB->up->down != NULL && LocalDB->down != NULL)
                LocalDB->up->down = LocalDB->down;
        }
        else
            DriverDatabase = NULL;
        RtlFreeHeap(GetProcessHeap(), 0, LocalDB->Driver);
    }
    return found;
}

/****************************************************************************
 *   find_driver(), finds a driver in the driver database.
 *
 * Parameters
 * object:           If not NULL, returns corrisponding database entry
 * key:              If not NULL, returns the corrisponding database entry;
 *                   if the previous parameter isn't NULL, this parameter is ignored
 * DeviceObject:     If not NULL, returns the corrisponding database entry, if the previous parameter isn't NULL,
 *                   this parameter is ignored
 * previous:         If not NULL, set to the PDEVICE_OBJECT previous to the DeviceObject parameter,
 *                   if DeviceObject is NULL, this parameter is ignored.
 * SymbolicLinkName: If not NULL, returns the corrisponding database entry. If all previous parameters
 *                   minus previous aren't NULL, this parameter is ignored.
 *
 * Return value:     A pointer to a db entry, see parameters for details..
 *                   If the correct database entry isn't found, NULL is returned.
 *                   If all parameters are NULL, NULL is returned.
 */
PDRIVERDATABASE find_driver(PDRIVER_OBJECT object, PWSTR key, PDEVICE_OBJECT DeviceObject,
                            PDEVICE_OBJECT previous, PUNICODE_STRING SymbolicLinkName)
{
    PDRIVERDATABASE LocalDB = DriverDatabase;
    int found = FALSE;
    
    if(object)
    {
        while(!found)
        {
            if(LocalDB->Driver == object)
            {
                found = TRUE;
                    break;
            }
            if(LocalDB->down != NULL)
                LocalDB = LocalDB->down;
            else
                    break;
        }
        goto done;
    }
    if(key)
    {
        while(!found)
        {
            if(lstrcmpiW(LocalDB->Key, key) == 0)
            {
                found = TRUE;
                break;
            }
            if(LocalDB->down != NULL)
                LocalDB = LocalDB->down;
            else
                    break;
        }
        goto done;
    }
    if(DeviceObject)
    {
        while(!found)
        {
            if(LocalDB->Driver->DeviceObject == DeviceObject)
            {
                found = TRUE;
                break;
            }
            if(LocalDB->Driver->DeviceObject->NextDevice != NULL && (previous))
            {
                previous = LocalDB->Driver->DeviceObject;
                LocalDB->Driver->DeviceObject = LocalDB->Driver->DeviceObject->NextDevice;
            }
            else
            {
                if(LocalDB->down != NULL)
                    LocalDB = LocalDB->down;
                else
                    break;
            }
        }
        goto done;
    }
    if(SymbolicLinkName)
    {
        while(!found)
        {
            if(!LocalDB->Driver)
                return NULL;
            if(RtlEqualUnicodeString(SymbolicLinkName,
            &(((DEVICE_NAME*)LocalDB->Driver->DeviceObject->Reserved)->DosDevicesName), TRUE))
            {
                found = TRUE;
                LocalDB = (PDRIVERDATABASE)LocalDB->Driver->DeviceObject;
                break;
            }
            if(LocalDB->Driver->DeviceObject->NextDevice)
            {
                LocalDB->Driver->DeviceObject = LocalDB->Driver->DeviceObject->NextDevice;
                continue;
            }
            if(LocalDB->down != NULL)
                LocalDB = LocalDB->down;
            else
                break;
        }
    }
    done:
    if(found == TRUE)
        return LocalDB;
    else
        return NULL;
}
